function [X, primal_cost, res] = solve_DRTR(objf, lambda, Y, evalf, opts)
% This function is adapted from the following function written by Bamdev Mishra and Gilles Meyer
% [model, preds_ts, infos] = lowrank_matrix_completion(algofun, data_ts, data_ls, model, params)
% We disabled warm start and regularization path
% Duality gap can be optionally switched on
%
% Computes the regularization path using a predictor-corrector approach.
% See the tech report http://arxiv.org/abs/1112.2318.
% Once lambda (the regularization parameter) is fixed, the optimization is
% done using low-rank optimization. All iterates are low-rank. Finally,
% optimality ceritificat is provided by computing the duality gap.
%
% Parameters
%
%   algofun                    function handle on the trust-region algorithm
%
%
%   data_ls.rows               vector containing first indices of the points
%   data_ls.cols               vector containing second indices of the points
%   data_ls.entries            vector of known entries
%   data_ts                    structure array for testing (if not empty)
%   (model.U,model.B,model.V)  initial point
%   model.gamma                reduction factor for creating a geometric
%                              sequence of regularizatin parameters
%                              starting from model.lambda_max
%   params                     structure array containing algorithm parameters
%                              (see default_arams for details)
%
% Output
%
%   model        structure array of low-rank polar factors
%   preds_ts     prediction for testing (computed only if data_ts is not empty)
%   infos        structure array with additional information
%
%
% Authors:
% Bamdev Mishra and Gilles Meyer
% {b.mishra, g.meyer}@ulg.ac.be

global tmp_obj_list tmp_loss_list tmp_perf_list tmp_time_list t1 total_time

gvalue = get_global_value();
col_print = gvalue.col_print;

algofun = @tr_obj;
opts.model.p = max(1, opts.model.p);

model = opts.model;
data_ls = [];
params = opts.params;
params = default_params(params);

max_iter = params.max_iter;
rtol = params.rtol;

verb = params.verb;
sigma_armijo = params.sigma_armijo;
max_step = params.max_step;
ls_max_iter = params.ls_max_iter;

smax_tol = params.smax_tol;
vp_tol = params.vp_tol;

dg_tol = params.dg_tol;
dg_vtol = params.dg_vtol;

d1 = model.d1;
d2 = model.d2;
p0 = model.p;
pmax = model.pmax;

lambda_max = lambda;
lambda_min = lambda;
gamma = model.gamma;

if ~isfield(model,'U') || ~isfield(model,'B') || ~isfield(model,'V'),
    model.U = uf(randn(d1,model.p));
    model.B = eye(model.p);
    model.V = uf(randn(d2,model.p));
end

res = [];
res.obj_list = [];
res.time_list = [];
res.loss_list = [];
res.perf_list = [];
tmp_obj_list = [];
tmp_loss_list = [];
tmp_perf_list = [];
tmp_time_list = [];

total_time = 0;
k = 1; % is a referencing variable
msg = [];
model.lambda = lambda_max;
model.num_flag = false;

break_flag = 0;
par = opts.ori_par;
msg = 'Running';

%   infos.costs = [];
%   infos.newRank = [];
%   infos.newRank_lambda = [];
%   infos.Regpath_sig_lambda = [];
%   infos.Regpath_duality_gap = [];
%   infos.Regpath_rel_duality_gap = [];
%   infos.Regpath_rank = [];
%   infos.Regpath_lambda = [];
%   infos.Regpath_sol = [];
%   infos.Regpath_warm_preds = [];
%   infos.Regpath_predictor_preds = [];
%   infos.Regpath_preds_ts = [];



%% Loop over lambda
while (model.lambda <= lambda_max && model.lambda >= lambda_min), % Lambda loop begins here
    
    token = 1; % A token to start optimization unless already initial point is optimal
    p0 = model.p;
    fprintf('######################  lambda = %g   ####################### \n',model.lambda);
    fprintf('*************\n');
    fprintf('*           *\n');
    fprintf('*  p = %i    *\n',model.p);
    fprintf('*           *\n');
    fprintf('*************\n');
    
%     warm_cost = objf({model.U*model.B, model.V}) + model.lambda*trace(model.B);
%     infos.Regpath_warm_preds = [ infos.Regpath_warm_preds ; warm_cost];
        
    %% Predictor-corrector step
    if k >= 3 && (p_1 == p_2),
        L = chol(B_2);
        eta.B = L'*logm(L'\B_1/L)*L;
        eta.U = U_1 - U_2;
        eta.V = V_1 - V_2;
        
        ProjDir = project_horizontal_space(eta,model); % Project from Euclidean space Horizontal space
        
        ProjDir.U = -ProjDir.U;
        ProjDir.B = -ProjDir.B;
        ProjDir.V = -ProjDir.V;
        
        step = gamma; % This is a very good guess of the step size, refer the paper for an explanation
        Prediction_before = warm_cost;
        fprintf('..... Prediction for warm restart is %g\n',Prediction_before);
        test_model = model;
        
        for ii = 1: ls_max_iter            
            if (p_1 == 1)
                test_model.U = U_2 + step*ProjDir.U;
                test_model.V = V_2 + step*ProjDir.V;
                test_model.U = test_model.U/sqrt(test_model.U'*test_model.U);
                test_model.V = test_model.V/sqrt(test_model.V'*test_model.V);
                test_model.B = B_2*exp(step*ProjDir.B/B_2);
            else
                B_half = chol(full(B_2));
                test_model.B = B_half'*expm(B_half'\(step*ProjDir.B)/B_half)*B_half;
                test_model.U = uf(U_2 + step*ProjDir.U);
                test_model.V = uf(V_2 + step*ProjDir.V);
            end
            
            Prediction_after = objf({test_model.U*test_model.B, test_model.V}) + ...
                                    test_model.lambda*trace(test_model.B);
            Armijo = Prediction_after - Prediction_before;
            ProjDir_norm_square = functions_nsym_polar_geometry('metric', [], model, ProjDir, ProjDir);
            
            if (Armijo <= -0.5*step*ProjDir_norm_square) || (Armijo <= 1e-30),
                break;
            else
                step = step/2;
            end            
        end
        
        model = test_model; % test model validated
        fprintf('..... Prediction using predictor-corrector is %g, #step search required = %i and step = %g \n',Prediction_after,ii, step);        
    else        
        ;       
    end
    
%     obj = objf({model.U*model.B, model.V}) + model.lambda*trace(model.B);
%     infos.Regpath_predictor_preds = [infos.Regpath_predictor_preds ; obj];

    % Test if the given point is close to the optimal solution by
    % observing the duality gap

    if params.check_dual
        [loss, dual_var] = objf({model.U*model.B, model.V});        
        [~, sig, ~] = lansvd(@(X)cell_times_mat(dual_var,X), ...
                             @(X)cell_trans_times_mat(dual_var,X), d1, d2, 1);

        % Duality gap
        primal_cost = loss + model.lambda*trace(model.B);
        dual_candidate = cell_times_const(dual_var, min(1, model.lambda/sig));        
        fenchel_dual = objf(dual_candidate, 'fdual');
        dual_cost = -fenchel_dual;
        duality_gap = primal_cost - dual_cost;
        rel_duality_gap = duality_gap / abs(dual_cost);

        col_print('primal=%g, dual=%g, gap=%g, rel=%g %% \n', ...
                primal_cost, dual_cost, duality_gap, rel_duality_gap);
%         fprintf('primal=%g, dual=%g, gap=%g, rel=%g %%\n', ...
%                 primal_cost, dual_cost, duality_gap, rel_duality_gap);

        %if  sig - model.lambda <= smax_tol,
        %if duality_gap <= dg_tol || rel_duality_gap <= dg_vtol,
        if rel_duality_gap <= dg_vtol,
            token = 0; % point is optimal
            infos.costs = [infos.costs;primal_cost];
            fprintf('[0000] cost = %g\n', primal_cost);
            fprintf('-------------------------------------------------\n');
            fprintf('OPTIMALITY CERTIFICATE with Relative duality gap %g \n',rel_duality_gap);
            fprintf('Duality gap is %g \n',duality_gap);
            fprintf('sigma - lambda = %g\n',sig - model.lambda);
            fprintf('-------------------------------------------------\n');
        end
    end
    
    %% Solvin the Trace norm minimization for a fixed lambda
    if token == 1
        for iter = 1:max_iter % If p = min(d1,d2), a global min is attained for sure
            
            t1 = cputime;
            
            if verb,
                fprintf('>> Rank %d <<\n', model.p);
            end
            
            if (model.p > p0),
                
                if isempty(restartDir), % No restart dir avail. do random restart
                    if verb,
                        disp('No restart dir available, do random restart');
                    end
                    
                    model.U = uf(randn(d1,model.p));
                    model.B = eye(model.p);
                    model.V = uf(randn(d2,model.p));
                    
                else % Perform line-search based on the restart direction
                    
                    if verb,
                        disp('>> Line-search with restart direction');
                    end
                    
                    costBefore = objf({model.U*model.B, model.V}) + model.lambda*trace(model.B);
                    if verb,
                        fprintf('>> Cost before = %f\n', costBefore);
                    end
                    
                    
                    beta = sig - model.lambda; % a good guess
                    
                    for i = 1:ls_max_iter,
                        
                        % Rank one update of polar factorization
                        test_model = svd_rank_1_update(model,beta*restartDir.U,restartDir.V);

                        costAfter = objf({test_model.U*test_model.B, test_model.V}) + ...
                                        test_model.lambda*trace(test_model.B);
                        if verb,
                            fprintf('>> Cost after = %f\n',costAfter);
                        end
                        
                        % Armijo condition
                        armijo = (costAfter - costBefore) <= 0;
                        if armijo,
                            break;
                        else
                            beta = beta / 2;
                        end
                        
                    end
                    model.U = test_model.U ; model.B = test_model.B ; model.V = test_model.V;
                    
                    
                    % Check for sufficient decrease
                    if (costAfter >= costBefore) || abs(costAfter - costBefore) < 1e-8,
                        
                        if verb,
                            disp('Decrease is not sufficient, random restart');
                        end
                        
                        model.U = uf(randn(d1,model.p));
                        model.B = eye(model.p);
                        model.V = uf(randn(d2,model.p));                        
                    end                    
                end                
            end
            
            %% Trust-region algorithm for fixed rank           
            [model, ignore, infos_algo] = feval(algofun, objf, evalf, [], data_ls, model, params);
            if model.num_flag
              msg = 'Stop with nan or inf';
              break;
            end
            
            %     	infos.costs = [infos.costs;infos_algo.costs];
            
            %% Dual variable candidate
            [loss, dual_var] = objf({model.U*model.B, model.V});
            primal_cost =  loss + model.lambda*trace(model.B);
            [u, sig, v] = lansvd(@(X)cell_times_mat(dual_var,X), ...
                                @(X)cell_trans_times_mat(dual_var,X), d1, d2, 1);
            base_time = total_time;
            total_time = total_time + (cputime - t1);
            
            obj = primal_cost;
            perf = evalf({model.U*model.B, model.V});

            if params.check_dual
                %% Duality gap computation
                dual_candidate = cell_times_const(dual_var, min(1, model.lambda/sig));        
                fenchel_dual = objf(dual_candidate, 'fdual');
                dual_cost = - fenchel_dual;
                duality_gap = primal_cost - dual_cost;
                rel_duality_gap = duality_gap/abs(dual_cost);

                col_print('%d: primal=%g, dual=%g, gap=%g, rel= %g %%, perf = (%g, %g), time: %g \n', iter, ...
                         primal_cost, dual_cost, duality_gap, rel_duality_gap*100, perf.l1, perf.l2, total_time);
%                 fprintf('%d: primal=%g, dual=%g, gap=%g, rel= %g %%\n', iter, ...
%                          primal_cost, dual_cost, duality_gap, rel_duality_gap*100);
            else
                col_print('%d: primal=%g, perf = (%g, %g), time: %g\n', ...
                            iter, primal_cost, perf.l1, perf.l2, total_time);
            end
            %     	if verb,
            %       		fprintf('>> sig_max - lambda = %g\n', sig - model.lambda);
            %       		fprintf('>> Relative duality gap = %g\n', rel_duality_gap);
            %       		fprintf('>> Duality gap = %g\n', duality_gap);
            %       end
                                    
            %% Stopping criterion            
            %     	%if sig - model.lambda <= smax_tol,
            %     	%if duality_gap <= dg_tol || rel_duality_gap <= dg_vtol,
            %     	%if duality_gap <= dg_tol,
            if params.check_dual && iter > 1 && rel_duality_gap <= dg_vtol,
                fprintf('-------------------------------------------------\n');
                fprintf('OPTIMALITY CERTIFICATE with Relative duality gap %g \n',rel_duality_gap);
                fprintf('Duality gap is %g \n', duality_gap);
                fprintf('sigma - lambda = %g\n', sig - model.lambda);
                fprintf('-------------------------------------------------\n');
                msg = 'Stop with small relative change of duality gap';
                break_flag = 1;
            end
            if iter > 1 && abs(pre_obj-obj) / min(abs([pre_obj, obj])) < rtol
                msg = 'Stop with small relative change of obj';
                break_flag = 1;
            elseif total_time > opts.max_time
                msg = 'Stop by max_time';
                break_flag = 1;
            elseif iter == max_iter
                msg = 'Stop with max iteration';
                break_flag = 1;
            else
                pre_obj = obj;
            end
            
%             res.obj_list = [res.obj_list; tmp_obj_list; obj];
%             res.loss_list = [res.loss_list; tmp_loss_list; loss];
%             res.perf_list = [res.perf_list; tmp_perf_list; perf];                
%             res.time_list = [res.time_list; tmp_time_list; total_time];
            res.obj_list = [res.obj_list; obj];
            res.loss_list = [res.loss_list; loss];
            res.perf_list = [res.perf_list; perf];                
            res.time_list = [res.time_list; total_time];
            res.msg = msg;
            res.iter = iter;
            
            tmp_obj_list = [];
            tmp_loss_list = [];
            tmp_perf_list = [];
            tmp_time_list = [];

            if isfield(opts, 'res_fname')
              save(opts.res_fname, 'res', 'par');
            end
            
            if break_flag == 1, break;    end
            
            %% Update rank and identify descent directions
            
            model.p = model.p + 1;
            
            if (sig > 1e-10),
                restartDir.U = -u;
                restartDir.V = v;
            else
                restartDir = [];
            end            
        end
    end
    
    %% Collect the information obtained
    
    %   infos.Regpath_duality_gap = [infos.Regpath_duality_gap; duality_gap];
    %   infos.Regpath_rel_duality_gap = [infos.Regpath_rel_duality_gap; rel_duality_gap];
    %   infos.Regpath_sig_lambda = [infos.Regpath_sig_lambda ; sig - model.lambda];
    %   infos.Regpath_rank = [infos.Regpath_rank; model.p];
    %   infos.Regpath_lambda = [infos.Regpath_lambda; model.lambda];
    %   % opt_solution = objf({model.U*model.B, model.V}) + model.lambda*trace(model.B);
    %   infos.Regpath_sol = [infos.Regpath_sol ; opt_solution];
    %
    % if (k >1) && (infos.Regpath_rank(end) > infos.Regpath_rank(end-1)),
    %     infos.newRank = [infos.newRank ; infos.Regpath_rank(end)];
    %     infos.newRank_lambda = [infos.newRank_lambda ; model.lambda ];
    %
    % end
    
    %% Data referencing
    
    %         if k == 1
    %             U_1 = model.U;
    %             V_1 = model.V;
    %             B_1 = model.B;
    %             p_1 = size(model.U,2);
    %         end;
    %         if k == 2;
    %             U_2 = model.U;
    %             V_2 = model.V;
    %             B_2 = model.B;
    %             p_2 = size(model.U,2);
    %         end;
    %
    %         if k >= 3
    %             U_1 = U_2;
    %             B_1 = B_2;
    %             V_1 = V_2;
    %             p_1 = p_2;
    %
    %             U_2 = model.U;
    %             V_2 = model.V;
    %             B_2 = model.B;
    %             p_2 = size(U_2, 2);
    %         end;
    
    k = k + 1;
    
    %% Update lambda
    model.lambda = gamma*model.lambda;
    %% If interested in computing prediction
%     if isfield(data_ts,'rows') && isfield(data_ts,'cols'),
%         preds_ts = proj_omega_sub(model.U*model.B, model.V, data_ts.rows, data_ts.cols);
%         errors = (preds_ts - data_ts.entries);
%         infos.Regpath_preds_ts =[infos.Regpath_preds_ts ; sum(errors.^2)];
%     else
%         preds_ts = [];
%     end
    
end % Lambda loops ends

if model.num_flag
  iter = iter - 1;
  X = [];
else  
  X = {model.U, model.B, model.V'};
end


if verb
  fprintf(1, '%s\n', msg);
end

end

