% [X, obj, iter, msg, outputs] = solve_SDP(fun, B, V, evalf, opts)
% solves the following problem:
%     min_X f(X) s.t. tr(X) <= B, X is positive semi-definite
% objf   : function handle
%            fun({V, V'}) should return f(V*V') and the gradient
% V     : Initial value (now ignored and just used to get the size of the matrices)
% lstype: line search type,   default: 2
%         0: b = 2/(k+2), a = 1-2/(k+2)
%         1: line search b, a = 1-2/(k+2)
%         2: line search a*V*V' + (1-a)*b*u*u', where a and b are locally searched in [0,1]
%        -1: line search both a and b subject to a+b<=1 (by fmincon)
% B     : trace radius

function [sol, obj, res] = solve_SDP(objf, B, V, evalf, opts)

opts = parse_param(opts);

if isempty(B),  B = 1;  end
if isinf(B) && ~opts.lstype   % must perform line search    
  opts.lstype = 2;
end

msg = 'Running';
break_flag = 0;
par = opts.ori_par;
  
res = struct('obj_list', [], ...
             'loss_list', [], ...
             'perf_list', [], ...
             'time_list', [] ...
             );

[d1, d2] = size(V);
m = d1 + d2;
V = randn(m, opts.init_rank);   % V has m rows

perf = 0;
total_time = 0;
for iter = 1:opts.max_iter
    
  t1 = cputime;

  [pre_obj, G] = objf({V(1:d1,:), V(d1+1:end,:)});        
  [u, ~] = laneig(@(x)cell_times_mat_SDP(G, x), m, 1, 'AS');
  if ~isinf(B)
    u = sqrt(B)*u;
  end

  [a, b, obj] = linesearch(V, u, B, iter, opts.lstype);

  V = [sqrt(a)*V, sqrt(b)*u];

  if ~isempty(evalf)
    perf = evalf({V(1:d1,:), V(d1+1:end,:)});
  end
  
  t1 = cputime - t1;  total_time = total_time + t1;

  if opts.verbose
    fprintf(1, 'iter: %d, a = %g, b = %g, loss = %g, perf = (%g, %g), time = %g\n', ...
      iter, a, b, obj, perf.l1, perf.l2, total_time);
  end

  if abs(obj - pre_obj) <= min(abs([obj, pre_obj]))*opts.rtol
    msg = 'Stop with small relative change';
    break_flag = 1;
  elseif isfield(opts, 'max_time') && total_time > opts.max_time
    msg = 'Stop by max_time';
    break_flag = 1;  
  elseif iter == opts.max_iter
    msg = 'Stop with max iteration';
    break_flag = 1;
  end  
  
  res.obj_list = [res.obj_list; obj];
  res.time_list = [res.time_list; total_time];
  res.loss_list = [res.loss_list; obj];
  res.perf_list = [res.perf_list; perf];
  res.msg = msg;
  res.iter = iter;
  
  if isfield(opts, 'res_fname')
    save(opts.res_fname, 'res', 'par');
  end
    
  if break_flag == 1, break;    end
end

sol = {V(1:d1,:), V(d1+1:end,:)};

if opts.verbose
  fprintf(1, '%s\n', msg);
end

  function [a, b, obj] = linesearch(V, u, B, ii, lstype)

    ab0 = [1/2; 1/2];
    if lstype == 0      % b = 2/(k+2), a = 1-2/(k+2)
      a = 1 - 2/(ii+2);
      b = 2/(ii+2);
      V = [sqrt(a)*V, sqrt(b)*u];
      obj = objf({V(1:d1,:), V(d1+1:end,:)});
    elseif lstype == 1  % line search b, a = 1-2/(k+2)
      a = 1 - 2/(ii+2);
      lsfun = @(b) gradb(a, b, V, u);
      [b, obj, it, nc, fl] = lbfgsb(ab0(2), 0, 2/(ii+2), lsfun);
    elseif lstype == 2  % line search both a and b, st a+b<=1, but locally
      if isinf(B)
        lsfun = @(ab) gradabinf(ab, V, u);
        [ab, obj, it, nc, fl] = lbfgsb(ab0, [0;0], Inf(2,1), lsfun);
        a = ab(1); b = ab(2);
      else
        lsfun = @(ab) gradab(ab, V, u);
        [ab, obj, it, nc, fl] = lbfgsb(ab0, [0;0], [1;1], lsfun);
        a = ab(1); b = (1-ab(1))*ab(2);
      end
    elseif lstype == -1 % line search both a and b subject to a+b<=1
      if isinf(B)
        lsfun = @(ab) gradabinf(ab, V, u);
        [ab, obj, it, nc, fl] = lbfgsb(ab0, [0;0], Inf(2,1), lsfun);
        a = ab(1); b = ab(2);
      else
        lsfun = @(ab) gradabinf(ab, V, u);
        opt = optimset('TolFun',1e-5, 'LargeScale' , 'off', 'Algorithm', ...
                        'interior-point', 'display', 'off');
        [ab, obj, fl] = fmincon(lsfun,[1/3;1/3],[1 1],1,[],[],[0;0],[1;1],[],opt);
        a = ab(1); b = ab(2);                
      end
    end
  end

  function [f, gb] = gradb(a, b, V, u)
    % f(a*V*V' + (1-a)*b*u*u'), where a is fixed
    VV = [sqrt(a)*V, sqrt(b)*u];
    [f, G] = objf({VV(1:d1,:), VV(d1+1:end,:)});
    gb = u' * cell_times_mat_SDP(G, u);
  end

  function [f, gab] = gradab(ab, V, u)
    % f(a*V*V' + (1-a)*b*u*u'), where a and b are locally searched in [0,1]
    VV = [sqrt(ab(1))*V, sqrt((1-ab(1))*ab(2))*u];
    [f, G] = objf({VV(1:d1,:), VV(d1+1:end,:)});
    gu = u' * cell_times_mat(G, u);
    gV = sum(sum(V .* cell_times_mat_SDP(G, V)));
    gab(1) = gV - ab(2)*gu;
    gab(2) = (1-ab(1))*gu;
  end

  function [f, gab] = gradabinf(ab, V, u)
    % f(a*V*V' + b*u*u'), a >= 0, b >= 0
    VV = [sqrt(ab(1))*V, sqrt(ab(2))*u];
    [f, G] = objf({VV(1:d1,:), VV(d1+1:end,:)});
    gab(1) = sum(sum(V .* cell_times_mat_SDP(G, V)));
    gab(2) = u' * cell_times_mat_SDP(G, u);
  end

  function opt_out = parse_param(opts_in)
    default = [];
    default.lstype = 2;
    default.max_iter = 1000;
    default.rtol = 1e-5;
    default.init_rank = 1;
    default.verbose = 1;
    opt_out = get_options(opts_in, default);
  end

  function [v] = cell_times_mat_SDP(G, x)
    v = [cell_times_mat(G, x(d1+1:end,:))
         cell_trans_times_mat(G, x(1:d1,:))];
  end

end
