function [w, f, res] = gcg_vec(loss, po, reg, w0, par)
% Inputs:
%   loss: the loss function handle which can be called by [f, g] = loss(x);
%   po: the polar oracle: 
%     Prototype:
%       [v, supp, fA] = pola(g);
%       Input:
%         G: the gradient to invoke the polar oracle
%         opt: options passed to the polar oracle
%              opt.weight_reg: the weight used in the prox operator
%       Output:
%         v: pola(g) with 0's in the complement of C(g) ignored
%         supp: reg(Z), value of regularizer at Z (not multiplied by weight_reg)
%         fA: F(C(g))^{1/q}
%   reg: the function handle that computes the regularization value
%   w0: initial solution
%   par: parameters
% 
% Outputs:
%   w: the optimal solution
%   f: the optimal objective value
  
  n = length(w0);
  w = zeros(n, 1);
  num_atom = 0;
  x = [];   % u^i is represented as a vector to allow direct call of lbfgsb
  s = 0;
  exp_par = par.exp_par;
  cg_par = par.cg_par;
  cone_par = par.cone_par;
  local_par = par.local_par;
  polar_par = par.polar_par;
  
  lambda = exp_par.lambda;
  max_iter = cg_par.max_cg_iter;
  
  % use x, supp_list, and fA to record the bases
  supp_list = cell(max_iter, 1);    % Support of each atom
  fA = zeros(max_iter, 1);          % F(A)^{1/q} of each atom
  bases = cell(max_iter, 1);        % Value of each atom
  s_list = zeros(max_iter, 1);      % weight on each atom
  s = 0;
  
  zero_const = zeros(max_iter,1);
  res = struct('flist', zero_const, ...
               'flist_ori', zero_const);
  rtime = struct('tm_all', zero_const, ...
                 'tm_loss', zero_const, ...
                 'tm_po', zero_const, ...
                 'tm_conic', zero_const, ...
                 'tm_local', zero_const);
    
  for i = 1 : max_iter-1
    
    reg_val = reg(w);
    
    tstart = cputime;
    [f, g] = loss.obj(w);
    rtime.tm_loss(i) = cputime - tstart;
    
    f_ori = f + lambda*reg_val;
    res.flist_ori(i) = f_ori;

    num_atom = num_atom + 1;
    t1 = cputime;
    [bases{num_atom}, supp_list{num_atom}, fA(num_atom), obj] = ...
                 po(-g, par); % v is compact with no zero

    if isfield(par, 'backup_pola') %
      [vv, ss, ffA, obj2] = par.backup_pola(-g, par); % v is compact with no zero
      cprintf('red', 'obj_sel = %g, obj_backup = %g\n', obj, obj2);
    end    
    rtime.tm_po(i) = cputime - t1;
        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Conic search
    t1 = cputime;
    switch cg_par.conic_search
      case '2d'
        if ~cg_par.use_local % use buffer for conic optimization
          loss.ctx = loss.update_ctx_2d_pre(loss.ctx, bases{num_atom}, supp_list{num_atom});
          [ab, f, iter, feval, flag] = lbfgsb([1;0], [0;0], [inf;inf], ...
                                              @obj_cone_2d_buf, [], ...
                                              cone_par.monitor, cone_par);
          loss.ctx = loss.update_ctx_2d_post(loss.ctx, ab);
        else
          [ab, f, iter, feval, flag] = lbfgsb([1;0], [0;0], [inf;inf], ...
                                              @obj_cone_2d, [], ...
                                              cone_par.monitor, cone_par);
        end
        s = [s, 1]*ab;
        s_list(1:num_atom) = [s_list(1:num_atom-1)*ab(1); ab(2)];
        w = ab(1)*w;
        supp = supp_list{num_atom};
        w(supp) = w(supp) + ab(2)*bases{num_atom};
        
        if cone_par.print_summary
          cprintf('red', '%d: cone: f=%f, iter=%d, feval=%d, a=%f, b=%f, msg=%s\n', ...
                  i, f, iter, feval, ab(1), ab(2), msg2str(flag, 'lbfgsb'));
        end
      case 'total'
        % Prepare loss context
        if ~cg_par.use_local % use buffer for conic optimization
          loss.ctx = loss.update_ctx_total(loss.ctx, bases{num_atom}, supp_list{num_atom});
          [wgts, f, iter, feval, flag] = lbfgsb([s_list(1:num_atom-1); 0], ...
                                              zeros(num_atom, 1), Inf(num_atom, 1), ...
                                              @obj_cone_total_buf, [], ...
                                              cone_par.monitor, cone_par);        
        else
          [wgts, f, iter, feval, flag] = lbfgsb([ones(num_atom-1,1);0], ...
                                            zeros(num_atom, 1), Inf(num_atom, 1), ...
                                            @obj_cone_total, [], ...
                                            cone_par.monitor, cone_par);
        end          
        s = sum(wgts);
        s_list(1:num_atom) = wgts;
        w = zeros(n,1);
        for k = 1 : num_atom
          supp = supp_list{k};
          w(supp) = w(supp) + wgts(k)*bases{k};
        end
        
        if cone_par.print_summary
          cprintf('red', '%d: cone: f=%f, iter=%d, feval=%d, msg=%s\n', ...
                    i, f, iter, feval, msg2str(flag, 'lbfgsb'));
        end

      case 'const'
        % use 1/k fixed updates
        b = 1/(1+i);  a = 1 - b;
        
        s = s*a + b;
        s_list(1:num_atom-1) = a*s_list(1:num_atom-1);
        s_list(num_atom) = b;
        
        w = a*w;
        supp = supp_list{num_atom};
        w(supp) = w(supp) + b*bases{num_atom};
    end
    rtime.tm_conic(i) = cputime - t1;

    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % local optimization
    if cg_par.use_local
      bu = Inf(length(x), 1);
      t1 = cputime;
      idx = 1;
      for k = 1:num_atom
        newidx = idx + length(supp_list{k}) - 1;
        x(idx:newidx) = bases{k} * s_list(k);
        idx = newidx+1;
      end
      [x, f, iter, feval, flag] = lbfgsb(x, -bu, bu, @obj_local, [], ...
                                        local_par.monitor, local_par);
      rtime.tm_local(i) = cputime - t1;
%       if local_par.print_summary
        cprintf('red', '%d: local: f=%f, iter=%d, feval=%d, msg=%s\n', ...
                i, f, iter, feval, msg2str(flag, 'lbfgsb'));    
%       end

      % Recover the basis representation of the solution
      w = zeros(n,1);
      s = 0;  idx = 1;
      for k = 1:num_atom
        supp = supp_list{k};
        newidx = idx + length(supp) - 1;
        xx = x(idx:newidx);        
        w(supp) = w(supp) + xx;

        pnm = pnorm(xx);
        bases{k} = xx / (fA(k)*pnm);
        s_list(k) = fA(k)*pnm;
        s = s + s_list(k);
        idx = newidx+1;
      end      
    end
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    rtime.tm_all(i) = cputime - tstart;
    
    % Update sum of weights
    res.flist(i) = f;
    
    if cg_par.verbose > 1
      cprintf('blue', 'iter = %d, f = %g, f_ori = %g, time = %g\n', ...
                    i, f, f_ori, sum(rtime.tm_all));
    end

    if sum(rtime.tm_all) > exp_par.max_time
      break;
    end
  end
  
  if cg_par.verbose > 0
    % Print summary of GCG
    ;
  end
    
  rtime.tm_all = rtime.tm_all(1:i);
  rtime.tm_loss = rtime.tm_loss(1:i);
  rtime.tm_po = rtime.tm_po(1:i);
  rtime.tm_conic = rtime.tm_conic(1:i);
  rtime.tm_local = rtime.tm_local(1:i);
  res.flist = res.flist(1:i);
  res.flist_ori = res.flist_ori(1:i);     
  res.rtime = rtime;
  res.iter = i;
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  % Convert a set of atoms into an original vector w
  function w = xlist_2_w(x)
    w = zeros(n, 1);
    idx = 1;
    for j = 1 : num_atom
      supp = supp_list{j};
      newidx = idx + length(supp) - 1;
      w(supp) = w(supp) + x(idx:newidx);
      idx = newidx + 1;
    end
  end

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Partially corrective
  % Objective of 2-D conic search
  function [f, g] = obj_cone_2d_buf(ab)
    [f, g] = loss.obj_cone_2d(ab, loss);
    
    f = f + [s,1]*ab*lambda;
    g = g + lambda*[s; 1];
  end

  function [f, g] = obj_cone_2d(ab)
    a = ab(1);  b = ab(2);
    cur_w = a*w;
    supp = supp_list{num_atom};
    cur_w(supp) = cur_w(supp) + b*bases{num_atom};
    [f, g] = loss.obj(cur_w);
    f = f + lambda*(a*s+b);
    g = g + lambda*[s; 1];
  end
  

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Totally corrective
  function [f, g] = obj_cone_total_buf(weights)    
    [f, g] = loss.obj_cone_total(weights, loss);
    f = f + lambda*sum(weights);
    g = g + lambda;
  end

  function [f, g] = obj_cone_total(weights)    
    g = zeros(num_atom, 1);
    cur_w = zeros(n, 1);
    idx = 1;
    for j = 1 : num_atom
      supp = supp_list{j};
      newidx = idx + length(supp) - 1;
      cur_w(supp) = cur_w(supp) + weights(j)*bases{j};      
      idx = newidx + 1;
    end
    
    [f1, g1] = loss.obj(cur_w);
    f = f1 + lambda*sum(weights);
    
    idx = 1;
    for j = 1 : num_atom-1
      supp = supp_list{j};
      newidx = idx + length(supp) - 1;
      g(j) = g1(supp)'*bases{j} + lambda;
      idx = newidx + 1;
    end
  end


  % Objective of local search
  function [f, g] = obj_local(x)
    [f1, g1] = loss.obj(xlist_2_w(x));
    
    f = 0;    g = zeros(length(x), 1);    idx = 1;
    for j = 1 : num_atom
      supp = supp_list{j};
      newidx = idx+length(supp)-1;
      tempx = x(idx:newidx);
      [f2, g2] = pnorm(tempx);
      f = f + fA(j)*f2;
      g(idx:newidx) = lambda*fA(j)*g2 + g1(supp);
      idx = newidx+1;
    end
    f = lambda*f + f1;
  end

  function [f, g] = pnorm(x)
    p = exp_par.p;
    
    if isinf(p) 
      % p = \infty, q = 1 (p is the ptil in the paper, q is the p in the paper).
      if cg_par.sigma < 0  % no smoothing
        [f, indmax] = max(abs(x));
        g = zeros(length(x),1); 
        g(indmax) = sign(x(indmax)) / length(indmax);
      else % use smoothing
        sigma = polar_par.sigma;
        y = abs(x);
        maxy = max(y);
        yexp = exp((y-maxy)/sigma);
        sumexp = sum(yexp);
        f = sigma*log(sumexp) + maxy;

        if nargout > 1
          g = yexp/sumexp .* sign(x);
        end
      end
    else  % never use smoothing for other norms
      y = abs(x);
      expx = y.^(p-1);
      sumexpx = sum(expx.*y);
      f = sumexpx^(1/p);
      if nargout > 1
        if f > 1e-6
          g = power(sumexpx, 1/p - 1)*(expx .* sign(x));
        else
          g = zeros(length(x),1);
        end
      end
    end
  end

end
